/*
 * Copyright (C) 2017 C-SKY Microsystems Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/******************************************************************************
 * @file     csi_ucos.c
 * @brief    the adapter file for the ucos
 * @version  V1.0
 * @date     20. July 2016
 ******************************************************************************/
#include <stdint.h>
#include <errno.h>
#include <csi_kernel.h>
#include <stddef.h>
#include <csi_config.h>
#include <soc.h>

/* ucos3 includes. */
#include <os.h>
#include <os_cfg.h>
#include <os_cfg_app.h>
#include <stdlib.h>
#include <string.h>

#include <mm.h>
#include <umm_heap.h>

#define OSSchedLockNestingCtr_MAX    (OS_NESTING_CTR)250
#define port_OS_MS_PERIOD_TICK       10
#define K_MSG_Q_SIZE                 128
#define TMR_ONE_SHOT_DLY             10
#define TMR_PERIODIC_PERIOD          10

static OS_STATE os_running_state(void)
{
    OS_STATE ret;
    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    ret = OSRunning;
    CPU_CRITICAL_EXIT();

    return ret;
}

k_status_t csi_kernel_init(void)
{
    OS_ERR err;

    OSInit(&err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_start(void)
{
    OS_ERR err;
    OSStart(&err);

    if (err == OS_ERR_FATAL_RETURN) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_sched_stat_t csi_kernel_get_stat(void)
{
    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return KSCHED_ST_READY;
    }

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();

    if (OSRunning == OS_STATE_OS_RUNNING) {
        if (OSSchedLockNestingCtr > (OS_NESTING_CTR)0) {
            CPU_CRITICAL_EXIT();
            return KSCHED_ST_LOCKED;
        } else {
            CPU_CRITICAL_EXIT();
            return KSCHED_ST_RUNNING;
        }
    } else if (OSRunning == OS_STATE_OS_STOPPED) {
        CPU_CRITICAL_EXIT();
        return KSCHED_ST_READY;
    }

    return KSCHED_ST_READY;
}

static int32_t get_SchedLockNestingCtr(void)
{
    int32_t ret = 0;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();

    if (OSSchedLockNestingCtr > (OS_NESTING_CTR)0) {
        ret = 1;
    }

    CPU_CRITICAL_EXIT();

    return ret;
}

static uint32_t get_LockNestingCtr(void)
{
    uint32_t ret;
    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    ret = OSSchedLockNestingCtr;
    CPU_CRITICAL_EXIT();
    return ret;
}

int32_t csi_kernel_sched_lock(void)
{
    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    OS_ERR err;
    int32_t  ret = get_LockNestingCtr();

    if (ret > 0) {
        ret = 1;
    } else if (ret < 0) {
        return -EPERM;
    }

    OSSchedLock(&err);

    if (err == OS_ERR_NONE) {
        return ret;
    } else {
        return -EPERM;
    }
}

int32_t csi_kernel_sched_unlock(void)
{
    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    OS_ERR err;
    int32_t  ret = get_LockNestingCtr();

    if (ret > 0) {
        ret = 1;
    } else if (ret < 0) {
        return -EPERM;
    }

    OSSchedUnlock(&err);

    if ((err == OS_ERR_NONE) || (err == OS_ERR_SCHED_LOCKED)) {
        return ret;
    } else {
        return -EPERM;
    }
}

static void set_SchedLockNestingCtr(uint32_t new_val)
{
    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    OSSchedLockNestingCtr = (OS_NESTING_CTR)new_val;
    CPU_CRITICAL_EXIT();
}

int32_t csi_kernel_sched_restore_lock(int32_t lock)
{
    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    if ((lock != 0) && (lock != 1)) {
        return -EINVAL;
    } else if (lock == 0) {
        set_SchedLockNestingCtr(0);
        return 0;
    } else if (lock == 1) {
        int32_t get_ctr = get_SchedLockNestingCtr();

        if (get_ctr == 1 || get_ctr == 0) {
            //Done csi_kernel_sched_unlock priviously, so increament SchedLockNestingCtr here.
            set_SchedLockNestingCtr(get_ctr + 1);
            return 1;
        } else if (get_ctr == OSSchedLockNestingCtr_MAX) {
            //Done csi_kernel_sched_lock priviously, so decreament SchedLockNestingCtr here, and it happens infrequently.
            set_SchedLockNestingCtr(get_ctr - 1);
            return 1;
        } else if (get_ctr >= 2 && get_ctr <= OSSchedLockNestingCtr_MAX - 1) {
            set_SchedLockNestingCtr(get_ctr + 1);
            return 1;
        }
    }

    return -EPERM;
}

uint32_t csi_kernel_sched_suspend(void)
{
    csi_kernel_sched_lock();
    return 0;
}

void csi_kernel_sched_resume(uint32_t sleep_ticks)
{
    csi_kernel_sched_unlock();
}

#if 0
static CPU_STK *csi_get_StkBasePtr(OS_TCB *ptr)
{
    return ptr->StkBasePtr;
}
#endif

k_status_t csi_kernel_task_new(k_task_entry_t task, const char *name, void *arg,
                               k_priority_t prio, uint32_t time_quanta,
                               void *stack, uint32_t stack_size, k_task_handle_t *task_handle)
{
    if ((stack_size % 4 != 0) || ((stack_size == 0) && (stack == NULL)) || prio == KPRIO_IDLE || prio > KPRIO_REALTIME7 || task_handle == NULL) {
        return -EINVAL;
    }

    OS_TCB *ret = (OS_TCB *)malloc(sizeof(OS_TCB));

    if (ret == NULL) {
        return -EPERM;
    }

    OS_ERR err;

    OS_PRIO prio_trans = (OS_CFG_PRIO_MAX - 1) - prio;
    uint32_t *stack_mem = NULL;
    stack_mem = csi_kernel_malloc(stack_size, __builtin_return_address(0));

    if (stack_mem == NULL) {
        free((void *)ret);
        return -EPERM;
    }

    csi_kernel_sched_lock();
    OSTaskCreate(ret, (char *)name, task, arg, prio_trans, stack_mem, 0, stack_size / 4, 0, time_quanta, NULL, OS_OPT_TASK_STK_CHK | OS_OPT_TASK_STK_CLR, &err);

    if (err == OS_ERR_NONE) {
        *task_handle = ret;
        csi_kernel_sched_unlock();
        return 0;
    } else {
        csi_kernel_sched_unlock();
        return -EPERM;
    }
}

//FIXME
CPU_STK *g_need_free_stack;
OS_TCB *g_need_free_handle;
k_status_t csi_kernel_task_del(k_task_handle_t task_handle)
{
    OS_ERR err;

    if (task_handle == NULL) {
        return -EINVAL;
    }

    csi_kernel_sched_suspend();

    if (g_need_free_handle) {
        free(g_need_free_handle);
        free(g_need_free_stack);
        g_need_free_handle = NULL;
        g_need_free_stack = NULL;
    }

    OS_TCB *get_handle = (OS_TCB *)task_handle;
    g_need_free_handle = get_handle;
    g_need_free_stack = get_handle->StkBasePtr;
    csi_kernel_sched_resume(0);

    if (csi_kernel_task_get_cur() == task_handle) {
        OSTaskDel(NULL, &err);
    } else {
        OSTaskDel(task_handle, &err);
    }

    return 0;
}

k_task_handle_t csi_kernel_task_get_cur(void)
{
    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return NULL;
    }

    OS_TCB *tsk_t;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    tsk_t = OSTCBCurPtr;
    CPU_CRITICAL_EXIT();

    return (k_task_handle_t)tsk_t;
}

k_task_stat_t csi_kernel_task_get_stat(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return KTASK_ST_ERROR;
    }

    if (csi_kernel_task_get_cur() == task_handle) {
        return KTASK_ST_RUNNING;
    }

    OS_STATE ret;
    OS_TCB *get_handle = (OS_TCB *)task_handle;
    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    ret = get_handle->TaskState;
    CPU_CRITICAL_EXIT();

    switch (ret) {
        case OS_TASK_STATE_RDY:
            return KTASK_ST_READY;
            break;

        case OS_TASK_STATE_DLY:
        case OS_TASK_STATE_PEND:
        case OS_TASK_STATE_PEND_TIMEOUT:
        case OS_TASK_STATE_SUSPENDED:
        case OS_TASK_STATE_DLY_SUSPENDED:
        case OS_TASK_STATE_PEND_SUSPENDED:
        case OS_TASK_STATE_PEND_TIMEOUT_SUSPENDED:
            return KTASK_ST_BLOCKED;
            break;

        case OS_TASK_STATE_DEL:
            return KTASK_ST_TERMINATED;
            break;

        default:
            return KTASK_ST_ERROR;
    }
}

k_status_t csi_kernel_task_set_prio(k_task_handle_t task_handle, k_priority_t priority)
{
    if (task_handle  == NULL || priority <= KPRIO_IDLE || priority > KPRIO_REALTIME7) {
        return -EINVAL;
    }

#if OS_CFG_TASK_CHANGE_PRIO_EN > 0u

    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    OS_ERR err;
    priority = (OS_CFG_PRIO_MAX - 1) - priority;
    OSTaskChangePrio(task_handle, priority, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EINVAL;
    }

#endif
}

k_priority_t csi_kernel_task_get_prio(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return KPRIO_ERROR;
    }

    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    OS_PRIO prio_get;
    OS_TCB *tsk = (OS_TCB *)task_handle;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    prio_get = tsk->Prio;
    CPU_CRITICAL_EXIT();

    prio_get = (OS_CFG_PRIO_MAX - 1) - prio_get;
    return prio_get;
}

const char *csi_kernel_task_get_name(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return NULL;
    }

    OS_TCB *tsk = (OS_TCB *)task_handle;

    return tsk->NamePtr;
}

k_status_t csi_kernel_task_suspend(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return -EINVAL;
    }

#if OS_CFG_TASK_SUSPEND_EN > 0u
    OS_ERR err;
    OSTaskSuspend(task_handle, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }

#endif
}

k_status_t csi_kernel_task_resume(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return -EINVAL;
    }

#if OS_CFG_TASK_SUSPEND_EN > 0u
    OS_ERR err;
    OSTaskResume(task_handle, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }

#endif
}

k_status_t csi_kernel_task_terminate(k_task_handle_t task_handle)
{
    return csi_kernel_task_del(task_handle);
}

void csi_kernel_task_exit(void)
{
    csi_kernel_task_del(csi_kernel_task_get_cur());
}

k_status_t csi_kernel_task_yield(void)
{
    return -EOPNOTSUPP;
}

static uint32_t get_OSTaskQty(void)
{
    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    uint32_t ret = 0;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    ret = OSTaskQty;
    CPU_CRITICAL_EXIT();

    return ret;
}

uint32_t csi_kernel_task_get_count(void)
{
    return get_OSTaskQty();
}

uint32_t csi_kernel_task_get_stack_size(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return 0;
    }

    OS_TCB *tsk = (OS_TCB *)task_handle;

    return 4 * (tsk->StkSize);
}

uint32_t csi_kernel_task_get_stack_space(k_task_handle_t task_handle)
{
    if (task_handle == NULL) {
        return 0;
    }

    OS_ERR err;
    CPU_STK_SIZE free, used;
    OSTaskStkChk(task_handle, &free, &used, &err);

    if (err == OS_ERR_NONE) {
        return free * 4;
    } else {
        return 0;
    }
}

uint32_t csi_kernel_task_list(k_task_handle_t *task_array, uint32_t array_items)
{
    //not support in ucos-iii so far.
    return 0;
}

k_status_t csi_kernel_intrpt_enter(void)
{
    OSIntEnter();
    return 0;
}

k_status_t csi_kernel_intrpt_exit(void)
{
    OSIntExit();
    return 0;
}

k_status_t csi_kernel_delay(uint32_t ticks)
{
    OS_ERR err;
    OSTimeDly(ticks, OS_OPT_TIME_DLY, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_delay_until(uint64_t ticks)
{
    if (ticks > (uint64_t) 0xffffffff) {
        return -EINVAL;
    }

    OS_ERR err;
    uint32_t cur = OSTimeGet(&err);

    if (cur >= ticks) {
        return 0;
    }

    OSTimeDly((uint32_t)ticks, OS_OPT_TIME_MATCH, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

uint64_t csi_kernel_tick2ms(uint32_t ticks)
{
    return ((uint64_t)ticks * port_OS_MS_PERIOD_TICK);
}

uint64_t csi_kernel_ms2tick(uint32_t ms)
{
    if (ms < port_OS_MS_PERIOD_TICK) {
        return 0;
    }

    return (((uint64_t)ms) / port_OS_MS_PERIOD_TICK);
}

k_status_t csi_kernel_delay_ms(uint32_t ms)
{
    uint32_t ms_get = ms;

    if ((ms != 0) && (ms < port_OS_MS_PERIOD_TICK)) {
        ms_get = port_OS_MS_PERIOD_TICK;
    }

    uint64_t ticks = csi_kernel_ms2tick(ms_get);
    OS_ERR err;

    OSTimeDly(ticks, OS_OPT_TIME_DLY, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

uint64_t csi_kernel_get_ticks(void)
{
    OS_ERR err;
    uint64_t get_tick = OSTimeGet(&err);

    if (err == OS_ERR_NONE) {
        return get_tick;
    }

    return 0;
}

void csi_kernel_update_tick(uint32_t ms)
{
}

uint32_t csi_kernel_get_tick_freq(void)
{
    return OS_CFG_TICK_RATE_HZ;
}

uint32_t csi_kernel_get_systimer_freq(void)
{
    return drv_get_sys_freq();
}

typedef struct tmr_ad {
    k_timer_cb_t   func;
    OS_TMR         tmr;
} tmr_ad_t;

typedef struct tmr_arg {
    void       *arg;
    tmr_ad_t   *tmr_above;
} tmr_arg_t;

static void tmr_adapt_cb(void *timer, void *arg)
{
    OS_TMR  *get = (OS_TMR *)timer;
    tmr_arg_t *arg_above = (tmr_arg_t *)(get->CallbackPtrArg);
    tmr_ad_t  *tmr_above = (tmr_ad_t *)arg_above->tmr_above;
    tmr_above->func(arg_above->arg);
    return;
}

k_timer_handle_t csi_kernel_timer_new(k_timer_cb_t func, k_timer_type_t type, void *arg)
{
    if (func == NULL) {
        return NULL;
    }

    OS_ERR err;

    tmr_ad_t *handle_ad = (tmr_ad_t *)malloc(sizeof(tmr_ad_t));

    if (handle_ad == NULL) {
        return NULL;
    }

    tmr_arg_t *get_arg = (tmr_arg_t *)malloc(sizeof(tmr_arg_t));

    if (get_arg == NULL) {
        free((void *)handle_ad);
        return NULL;
    }

    get_arg->arg = arg;

    OS_OPT opt;

    if (type == KTIMER_TYPE_ONCE) {
        opt = OS_OPT_TMR_ONE_SHOT;
    } else {
        opt = OS_OPT_TMR_PERIODIC;
    }

    OSTmrCreate(&(handle_ad->tmr), NULL, TMR_ONE_SHOT_DLY, TMR_PERIODIC_PERIOD, opt, (OS_TMR_CALLBACK_PTR)tmr_adapt_cb, get_arg, &err);

    if (err == OS_ERR_NONE) {
        handle_ad->func = func;
        get_arg->tmr_above = handle_ad;
        return handle_ad;
    } else {
        free((void *)handle_ad);
        free((void *)get_arg);
        return NULL;
    }
}

k_status_t csi_kernel_timer_del(k_timer_handle_t timer_handle)
{
    if (timer_handle == NULL) {
        return -EINVAL;
    }

    tmr_ad_t *handle_ad = (tmr_ad_t *)timer_handle;

    OS_ERR err;
    OSTmrDel(&(handle_ad->tmr), &err);

    if (err == OS_ERR_NONE) {
        free(((OS_TMR *)(&(handle_ad->tmr)))->CallbackPtrArg);
        free(timer_handle);
        return 0;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_timer_start(k_timer_handle_t timer_handle, uint32_t ticks)
{
    if (timer_handle == NULL || ticks == 0) {
        return -EINVAL;
    }

    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    OS_ERR err;
    tmr_ad_t *handle_ad = (tmr_ad_t *)timer_handle;
    OS_TMR *tmr = (OS_TMR *)(&(handle_ad->tmr));

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();

    if (tmr->Opt == OS_OPT_TMR_PERIODIC) {
        tmr->Dly = 0;
        tmr->Period = ticks;
    } else if (tmr->Opt == OS_OPT_TMR_ONE_SHOT) {
        tmr->Dly = ticks;
        tmr->Period = 0;
    }

    CPU_CRITICAL_EXIT();

    OSTmrStart(tmr, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_timer_stop(k_timer_handle_t timer_handle)
{
    if (timer_handle == NULL) {
        return -EINVAL;
    }

    tmr_ad_t *handle_ad = (tmr_ad_t *)timer_handle;
    OS_TMR *tmr = (OS_TMR *)(&(handle_ad->tmr));

    OS_ERR err;
    OSTmrStop(tmr, OS_OPT_TMR_NONE, NULL, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_timer_stat_t csi_kernel_timer_get_stat(k_timer_handle_t timer_handle)
{
    if (timer_handle == NULL) {
        return KTIMER_ST_INACTIVE;
    }

    tmr_ad_t *handle_ad = (tmr_ad_t *)timer_handle;
    OS_TMR *tmr = (OS_TMR *)(&(handle_ad->tmr));

    OS_ERR err;
    OS_STATE ret = OSTmrStateGet(tmr, &err);

    if (err == OS_ERR_NONE) {
        if (ret == OS_TMR_STATE_RUNNING) {
            return KTIMER_ST_ACTIVE;
        } else {
            return KTIMER_ST_INACTIVE;
        }
    }

    return KTIMER_ST_INACTIVE;
}

k_event_handle_t csi_kernel_event_new(void)
{
    OS_ERR err;
    OS_FLAG_GRP *ret = (OS_FLAG_GRP *)malloc(sizeof(OS_FLAG_GRP));

    if (ret == NULL) {
        return NULL;
    }

    OSFlagCreate(ret, NULL, 0, &err);

    if (err == OS_ERR_NONE) {
        return ret;
    } else {
        free(ret);
        return NULL;
    }
}

k_status_t csi_kernel_event_del(k_event_handle_t ev_handle)
{
    if (ev_handle == NULL) {
        return -EINVAL;
    }

#if OS_CFG_FLAG_DEL_EN > 0u
    OS_ERR err;
    OSFlagDel(ev_handle, OS_OPT_DEL_ALWAYS, &err);

    if (err == OS_ERR_NONE) {
        free(ev_handle);
        return 0;
    } else {
        return -EPERM;
    }

#endif
}

k_status_t csi_kernel_event_set(k_event_handle_t ev_handle, uint32_t flags, uint32_t *ret_flags)
{
    if (ev_handle == NULL || ret_flags == NULL) {
        return -EINVAL;
    }

    OS_ERR err;
    OS_FLAGS ret = OSFlagPost(ev_handle, flags, OS_OPT_POST_FLAG_SET, &err);

    if (err == OS_ERR_NONE) {
        *ret_flags = ret;
        return 0;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_event_clear(k_event_handle_t ev_handle, uint32_t flags, uint32_t *ret_flags)
{
    if (ev_handle == NULL || ret_flags == NULL) {
        return -EINVAL;
    }

    OS_FLAGS get;
    csi_kernel_sched_lock();
    k_status_t ret_1 = csi_kernel_event_get(ev_handle, &get);

    if (ret_1 != 0) {
        csi_kernel_sched_unlock();
        return -EPERM;
    }

    *ret_flags = get;

    OS_ERR err;
    OSFlagPost(ev_handle, flags, OS_OPT_POST_FLAG_CLR, &err);

    if (err == OS_ERR_NONE) {
        csi_kernel_sched_unlock();
        return 0;
    } else {
        csi_kernel_sched_unlock();
        return -EPERM;
    }
}

k_status_t csi_kernel_event_get(k_event_handle_t ev_handle, uint32_t *ret_flags)
{
    if (ev_handle == NULL || ret_flags == NULL) {
        return -EINVAL;
    }

    OS_FLAGS flags;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    flags = ((OS_FLAG_GRP *)ev_handle)->Flags;
    CPU_CRITICAL_EXIT();
    *ret_flags = flags;

    return 0;
}

k_status_t csi_kernel_event_wait(k_event_handle_t ev_handle, uint32_t flags,
                                 k_event_opt_t options, uint8_t clr_on_exit,
                                 uint32_t *actl_flags, int32_t timeout)
{
    if (ev_handle == NULL || actl_flags == NULL) {
        return -EINVAL;
    }

    OS_OPT opt_tmp = OS_OPT_PEND_BLOCKING;
    uint32_t clear_mark = 0;
    uint32_t setb_mark = 0;

    if (timeout < 0) {
        timeout = 0;
    } else if (timeout == 0) {
        opt_tmp = OS_OPT_PEND_NON_BLOCKING;
    }

    if (clr_on_exit == 1) {
        opt_tmp += OS_OPT_PEND_FLAG_CONSUME;
        clear_mark = 1;
    } else if (clr_on_exit != 0 && clr_on_exit != 1) {
        return -EINVAL;
    }

    switch (options) {
        case KEVENT_OPT_SET_ANY:
            setb_mark = 1;
            opt_tmp += OS_OPT_PEND_FLAG_SET_ANY;
            break;

        case KEVENT_OPT_SET_ALL:
            setb_mark = 1;
            opt_tmp += OS_OPT_PEND_FLAG_SET_ALL;
            break;

        case KEVENT_OPT_CLR_ANY:
            opt_tmp += OS_OPT_PEND_FLAG_CLR_ANY;
            break;

        case KEVENT_OPT_CLR_ALL:
            opt_tmp += OS_OPT_PEND_FLAG_CLR_ALL;
            break;

        default:
            return -EINVAL;
    }

    OS_ERR err;

    uint32_t pre_get;
    uint32_t after_get;
    csi_kernel_event_get(ev_handle, &pre_get);

    OS_FLAGS event_val = OSFlagPend(ev_handle, flags, timeout, opt_tmp, NULL, &err);

    csi_kernel_event_get(ev_handle, &after_get);

    if (err == OS_ERR_NONE) {
        OS_FLAGS ready_get = OSFlagPendGetFlagsRdy(&err);

        if (err != OS_ERR_NONE) {
            /*FIX ME: to do.*/
            return 0;
        }

        if (clear_mark) {
            if (setb_mark) {
                if (pre_get == event_val) {
                    *actl_flags = event_val | ready_get;
                    return 0;
                } else {
                    *actl_flags = pre_get | event_val | ready_get;
                    return 0;
                }
            }
        } else {
            *actl_flags = after_get;
            return 0;
        }

        return -EPERM;
    } else if (err == OS_ERR_TIMEOUT) {
        return -ETIMEDOUT;
    } else {
        return -EPERM;
    }
}

k_mutex_handle_t csi_kernel_mutex_new(void)
{
    OS_MUTEX *ret = (OS_MUTEX *)malloc(sizeof(OS_MUTEX));

    if (ret == NULL) {
        return NULL;
    }

    OS_ERR err;

    OSMutexCreate(ret, NULL, &err);

    if (err == OS_ERR_NONE) {
        return ret;
    } else {
        free(ret);
        return NULL;
    }
}

k_status_t csi_kernel_mutex_del(k_mutex_handle_t mutex_handle)
{
    if (mutex_handle == NULL) {
        return -EINVAL;
    }

#if OS_CFG_MUTEX_DEL_EN > 0u
    OS_ERR err;
    OSMutexDel(mutex_handle, OS_OPT_DEL_ALWAYS, &err);

    if (err == OS_ERR_NONE) {
        free(mutex_handle);
        return 0;
    } else {
        return -EPERM;
    }

#endif
}

k_status_t csi_kernel_mutex_lock(k_mutex_handle_t mutex_handle, int32_t timeout)
{
    if (mutex_handle == NULL) {
        return -EINVAL;
    }

    OS_OPT get_opt = OS_OPT_PEND_BLOCKING;

    if (timeout < 0) {
        timeout = 0;
    } else if (timeout == 0) {
        get_opt = OS_OPT_PEND_NON_BLOCKING;
    }

    OS_ERR err;
    OSMutexPend(mutex_handle, timeout, get_opt, NULL, &err);

    if (err == OS_ERR_NONE || err == OS_ERR_MUTEX_OWNER) {
        return 0;
    } else if (err == OS_ERR_TIMEOUT) {
        return -ETIMEDOUT;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_mutex_unlock(k_mutex_handle_t mutex_handle)
{
    if (mutex_handle == NULL) {
        return -EINVAL;
    }

    OS_ERR err;
    OSMutexPost(mutex_handle, OS_OPT_POST_NONE, &err);

    if (err == OS_ERR_NONE || err == OS_ERR_MUTEX_NESTING) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_task_handle_t csi_kernel_mutex_get_owner(k_mutex_handle_t mutex_handle)
{
    if (mutex_handle == NULL) {
        return NULL;
    }

    OS_MUTEX *mut = (OS_MUTEX *)mutex_handle;
    return mut->OwnerTCBPtr;
}

k_sem_handle_t csi_kernel_sem_new(int32_t max_count, int32_t initial_count)
{
    if (max_count <= 0 || initial_count < 0) {
        return NULL;
    }

    if (max_count < initial_count) {
        return NULL;
    }

    OS_SEM *ret = (OS_SEM *)malloc(sizeof(OS_SEM));

    if (ret == NULL) {
        return NULL;
    }

    OS_ERR err;

    OSSemCreate(ret, NULL, initial_count, &err);

    if (err == OS_ERR_NONE) {
        return ret;
    } else {
        free(ret);
        return NULL;
    }
}

k_status_t csi_kernel_sem_del(k_sem_handle_t sem_handle)
{
    if (sem_handle == NULL) {
        return -EINVAL;
    }

#if OS_CFG_SEM_DEL_EN > 0u
    OS_ERR err;
    OSSemDel(sem_handle, OS_OPT_DEL_ALWAYS, &err);

    if (err == OS_ERR_NONE) {
        free(sem_handle);
        return 0;
    } else {
        return -EPERM;
    }

#endif
}

k_status_t csi_kernel_sem_wait(k_sem_handle_t sem_handle, int32_t timeout)
{
    if (sem_handle == NULL) {
        return -EINVAL;
    }

    OS_OPT get_opt = OS_OPT_PEND_BLOCKING;

    if (timeout < 0) {
        timeout = 0;
    } else if (timeout == 0) {
        get_opt = OS_OPT_PEND_NON_BLOCKING;
    }

    OS_ERR err;
    OSSemPend(sem_handle, timeout, get_opt, NULL, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else if (err == OS_ERR_TIMEOUT) {
        return -ETIMEDOUT;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_sem_post(k_sem_handle_t sem_handle)
{
    if (sem_handle == NULL) {
        return -EINVAL;
    }

    OS_ERR err;
    OSSemPost(sem_handle, OS_OPT_POST_1, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

int32_t csi_kernel_sem_get_count(k_sem_handle_t sem_handle)
{
    if (sem_handle == NULL) {
        return -EINVAL;
    }

    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    uint32_t ret;
    OS_SEM *sem = (OS_SEM *)sem_handle;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    ret = sem->Ctr;
    CPU_CRITICAL_EXIT();

    return ret;
}

k_mpool_handle_t csi_kernel_mpool_new(void *p_addr, int32_t block_count, int32_t block_size)
{
    if (p_addr == NULL || block_count < 0 || block_size < 0 || (block_size % 4 != 0)) {
        return NULL;
    }

    OS_MEM *ret = (OS_MEM *)malloc(sizeof(OS_MEM));

    if (ret == NULL) {
        return NULL;
    }

    OS_ERR err;
    OSMemCreate(ret, NULL, p_addr, block_count, block_size, &err);

    if (err == OS_ERR_NONE) {
        return ret;
    } else {
        free(ret);
        return NULL;
    }
}

k_status_t csi_kernel_mpool_del(k_mpool_handle_t mp_handle)
{
    if (mp_handle == NULL) {
        return -EINVAL;
    }

    memset(mp_handle, 0, sizeof(OS_MEM));
    free(mp_handle);
    return 0;
}

void *csi_kernel_mpool_alloc(k_mpool_handle_t mp_handle, int32_t timeout)
{
    if (mp_handle == NULL) {
        return NULL;
    }

    uint32_t *ret = 0;
    OS_ERR err;
    ret = OSMemGet(mp_handle, &err);

    if (err == OS_ERR_NONE) {
        return (void *)ret;
    } else {
        return NULL;
    }
}

k_status_t csi_kernel_mpool_free(k_mpool_handle_t mp_handle, void *block)
{
    if (mp_handle == NULL || block == NULL) {
        return -EINVAL;
    }

    OS_ERR err;
    OSMemPut(mp_handle, block, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

int32_t csi_kernel_mpool_get_count(k_mpool_handle_t mp_handle)
{
    if (mp_handle == NULL) {
        return -EINVAL;
    }

    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    OS_MEM *blk = (OS_MEM *)mp_handle;
    OS_MEM_QTY ret = 0;

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    ret = blk->NbrMax - blk->NbrFree;
    CPU_CRITICAL_EXIT();

    return ret;
}

uint32_t csi_kernel_mpool_get_capacity(k_mpool_handle_t mp_handle)
{
    if (mp_handle == NULL) {
        return 0;
    }

    OS_MEM *blk = (OS_MEM *)mp_handle;
    return blk->NbrMax;
}

uint32_t csi_kernel_mpool_get_block_size(k_mpool_handle_t mp_handle)
{
    if (mp_handle == NULL) {
        return 0;
    }

    OS_MEM *blk = (OS_MEM *)mp_handle;
    return blk->BlkSize;
}

typedef struct mq_adapter {
    OS_Q    os_q;
    int32_t msg_size;
    int32_t msg_count;
} mq_adapter_t;
k_msgq_handle_t csi_kernel_msgq_new(int32_t msg_count, int32_t msg_size)
{
    if (msg_count <= 0 || msg_size <= 0) {
        return NULL;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)malloc(sizeof(mq_adapter_t));

    if (mq_adapter == NULL) {
        return NULL;
    }

    OS_ERR err;
    OSQCreate(&(mq_adapter->os_q), NULL, msg_count, &err);

    if (err != OS_ERR_NONE) {
        free(mq_adapter);
        return NULL;
    }

    mq_adapter->msg_size = msg_size;
    return (k_msgq_handle_t)mq_adapter;
}

k_status_t csi_kernel_msgq_del(k_msgq_handle_t mq_handle)
{
#if OS_CFG_Q_DEL_EN > 0u

    if (!mq_handle) {
        return -EINVAL;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;
    OS_ERR err;
    OSQDel(&(mq_adapter->os_q), OS_OPT_DEL_ALWAYS, &err);

    if (err == OS_ERR_NONE) {
        free(mq_handle);
        return 0;
    } else {
        return -EPERM;
    }

#endif
}

k_status_t csi_kernel_msgq_put(k_msgq_handle_t mq_handle, const void *msg_ptr, uint8_t front_or_back, int32_t timeout)
{
    if ((!mq_handle) || (msg_ptr == NULL) || ((front_or_back != 0) && (front_or_back != 1))) {
        return -EINVAL;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;
    int opt_get = 0;

    if (front_or_back == 0) {
        opt_get += OS_OPT_POST_FIFO;
    } else if (front_or_back == 1) {
        opt_get += OS_OPT_POST_LIFO;
    }

    OS_ERR err;
    OSQPost(&(mq_adapter->os_q), (void *)msg_ptr, mq_adapter->msg_size, opt_get, &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

k_status_t csi_kernel_msgq_get(k_msgq_handle_t mq_handle, void *msg_ptr, int32_t timeout)
{
    if (mq_handle == NULL || msg_ptr == NULL) {
        return -EINVAL;
    }

    OS_OPT opt_get = OS_OPT_PEND_BLOCKING;

    if (timeout < 0) {
        timeout = 0;
    } else if (timeout == 0) {
        opt_get = OS_OPT_PEND_NON_BLOCKING;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;
    OS_MSG_SIZE msgq_size;
    void *msg;

    OS_ERR err;
    msg = OSQPend(&(mq_adapter->os_q), timeout, opt_get, &msgq_size, NULL, &err);

    if ((err == OS_ERR_NONE) && (msg != NULL)) {
        memcpy(msg_ptr, msg, mq_adapter->msg_size);
        return 0;
    } else {
        return -EPERM;
    }
}

int32_t csi_kernel_msgq_get_count(k_msgq_handle_t mq_handle)
{
    if (mq_handle == NULL) {
        return -EINVAL;
    }

    if (os_running_state() == OS_STATE_OS_STOPPED) {
        return -EPERM;
    }

    uint32_t get_ctr = 0;
    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;

    OS_Q *get_mq = &(mq_adapter->os_q);
    OS_MSG_Q *p = (OS_MSG_Q *)(&(get_mq->MsgQ));

    CPU_SR_ALLOC();
    CPU_CRITICAL_ENTER();
    get_ctr = p->NbrEntries;
    CPU_CRITICAL_EXIT();

    return get_ctr;
}

uint32_t csi_kernel_msgq_get_capacity(k_msgq_handle_t mq_handle)
{
    if (mq_handle == NULL) {
        return 0;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;

    OS_Q *get_mq = &(mq_adapter->os_q);
    OS_MSG_Q *p = (OS_MSG_Q *)(&(get_mq->MsgQ));
    return p->NbrEntriesSize;
}

uint32_t csi_kernel_msgq_get_msg_size(k_msgq_handle_t mq_handle)
{
    if (mq_handle == NULL) {
        return 0;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;
    return mq_adapter->msg_size;
}

k_status_t csi_kernel_msgq_flush(k_msgq_handle_t mq_handle)
{
    if (mq_handle == NULL) {
        return -EINVAL;
    }

    mq_adapter_t *mq_adapter = (mq_adapter_t *)mq_handle;

    OS_ERR err;
    OSQFlush(&(mq_adapter->os_q), &err);

    if (err == OS_ERR_NONE) {
        return 0;
    } else {
        return -EPERM;
    }
}

void *csi_kernel_malloc(int32_t size, void *caller)
{
    void *ret;

    csi_kernel_sched_suspend();
    ret = mm_malloc(USR_HEAP, size, caller);
    csi_kernel_sched_resume(0);
    return ret;
}

void csi_kernel_free(void *ptr, void *caller)
{
    csi_kernel_sched_suspend();
    mm_free(USR_HEAP, ptr, caller);
    csi_kernel_sched_resume(0);
}

void *csi_kernel_realloc(void *ptr, int32_t size, void *caller)
{
    void *new_ptr;

    new_ptr = csi_kernel_malloc(size, caller);

    if (ptr) {
        memcpy(new_ptr, ptr, size);
        csi_kernel_free(ptr, caller);
    }

    return new_ptr;
}

k_status_t csi_kernel_get_mminfo(int32_t *total, int32_t *used, int32_t *free, int32_t *peak)
{
    mm_get_mallinfo(total, used, free, peak);

    return 0;
}

k_status_t csi_kernel_mm_dump(void)
{
    mm_leak_dump();

    return 0;
}
